#ifndef GENOTYPE_H_
#define GENOTYPE_H_
#include "individuDiscriminant.h"

FILE* result;
std::string  __seed;
/**
 * @class Gene : public eoPrintable
 * @file eoCPTEO_V1.h
 * @author Jacques Bibai
 * @brief definition d'une gare (station intermediaire) dans la metaphore du TGV.
 * Un gene represente une gare dans la metaphore du TGV. 
 * Ici il est defini comme un vecteur d'entiers. 
 * Chaque entier representant l'indice d'un fluent dans la table des fluents.
 * La table des fluents est creee lors de l'initialisation de l'objet global. 
 */
class Gene : public eoPrintable 
{
	/**
	 * Attribut de la classe Gene
	 */
	private :
		std::vector<size_t> gene; // liste d'indices de fluents
		size_t bk;
		
	public :
		/**
		 * Constructeur vide
		 */
		Gene()
		{
		}
		void setbk(size_t i){
			bk=i;
		}
		
		size_t getbk(void)
		{
			return bk;
		}
		/**
		 * @fn std::vector<size_t> & getGene()
		 * @brief Accesseur a l'attribut prive gene
		 */
		 
		std::vector<size_t> & getGene()
		{
			return gene;
		}
		
		/**
		 * @fn size_t size()
		 * @brief calcul de la taille du vecteur gene
		 * @return taille du vecteur gene
		 */
		size_t size()
		{
			return gene.size();
		}
		
		/**
		 * @fn void set(size_t, size_t)
		 * @brief Mise a jour d'un element de gene
		 * @param position : entier non signe
		 * @param valeur : nouvelle valeur (indice du fluent dans la table)
		 * remplace la valeur de gene[position] par valeur
		 */
		void set(size_t position, size_t valeur)
		{
			gene[position]=valeur;
		}
		
		/**
		 * @fn size_t get(size_t)
		 * @brief donne la valvaleur de la position :
		 * @param position : entier non signe representant la position de l'indice du fluent a recuperer
		 * @return entier non signe egal a la valeur de gene[position]
		 */
		size_t get(size_t position)
		{
			return gene[position];
		}
		
		/**
		 * @fn int mutex(int,size_t)
		 * @brief verification de l'exclusion mutuelle
		 * @param idfluent : indice du fluent a tester
		 * @param pos : position dans gene a sauter lors de la verification
		 * @return 1 s'il existe un indice dans gene mutuellement exclusif avec idfluent 0 sinon
		 * cette methode permet de verifier si fluent d'indice idfluent est mutuellement exclusif avec
		 * tous les fluents contenus dans gene excepte celui qui ce trouve a la position pos
		 * 
		 */
		int mutex(int idfluent,size_t pos)
		{
			if(pos==0)return 0;
			int test=0;
			for(size_t i=0;i<gene.size();i++){
				if(i!=pos)
					test=cptMutex(cptGetFluentFromIndex(gene[i]),cptGetFluentFromIndex(idfluent));
				if(test)break;
			}
			return test;
			
		}
		/**
		 * @fn int mutex2(int, size_t)
		 * @brief verification de l'exclusion mutuelle
		 * @param idfluent : indice du fluent a tester
		 * @param pos : indice de la table a ne pas depasser (nombre d'element de la table conserne par le test a partir de l'element se trouvant a la position 0)
		 * @return 1 s'il existe un indice dans gene mutuellement exclusif avec idfluent 0 sinon
		 * cette methode permet de verifier si fluent d'indice idfluent est mutuellement exclusif avec
		 * tous les fluents contenus dans gene de l'indice 0 a l'incide pos-1 
		 */
		int mutex2(int idfluent,size_t pos)
		{
			if(pos==0)return 0;
			int test=0;
			for(size_t i=0;i<pos;i++){
				test=cptMutex(cptGetFluentFromIndex(gene[i]),cptGetFluentFromIndex(idfluent));
				if(test)break;
			}
			return test;
			
		}
		
		/**
		 * @pre le fluent mutuellement exclusif doit exister
		 * @return le fluent mutuellement exclusif
		 */
		int mutex3(int idfluent,size_t pos)
		{
			if(pos==0)return 0;
			int test=0;size_t i=0;
			for(i=0;i<pos;i++){
				test=cptMutex(cptGetFluentFromIndex(gene[i]),cptGetFluentFromIndex(idfluent));
				if(test)break;
			}
			return gene[i];
			
		}
		
		/**
		 * @return la position d'un fluent donne ou -1 sinon
		 */
		int position(int idFluent){
			int i;
			for(i=0;(size_t)i<gene.size();i++){
				if((size_t)idFluent==(size_t)gene[i])return i;
			}
			return -1;
		}
		
		/**
		 * @return le fluent s'il existe et -1 sinon
		 */
		 int exist(int idFluent){
			int i;
			for(i=0;(size_t)i<gene.size();i++){
				if((size_t)idFluent==(size_t)gene[i])return gene[i];
			}
			return -1;
		}
		 int exist(int idFluent,int pos){
			int i;
			for(i=0;i<pos;i++){
				if((size_t)idFluent==(size_t)gene[i])return gene[i];
			}
			return -1;
		}
		/**
		 * @fn std::vector<string> toString()
		 * @brief transforme un gene en vecteur de chaine
		 * permet de convertir la table d'indices de fluent et table
		 * de chaine de characteres
		 */
		std::vector<string> toString()
		{
			std::vector<string> tmp;
			tmp.resize(gene.size());
			
			for(size_t i=0;i<gene.size();i++){
				tmp[i]= cptGetFluentName(cptGetFluentFromIndex(gene[i]));
				
			}
			return tmp;
		}
		
		/**
		 * @fn void printOn(ostream&)
		 * @brief methode printOn permettant d'afficher une instance de Gene
		 */		
		void printOn(ostream& os) const
		{
		  size_t i;
		  os << " "<< gene.size()<<" ";
		  for (i = 0; i < gene.size(); i++) {
		    
		     // affichage d'un fluent
		    os << " "<< gene[i] << " " <<cptGetFluentName(cptGetFluentFromIndex(gene[i])) << " "; 
			}
			os << "BK " << bk;
		}
		
		void printCPT(FILE* out)const
		{
			size_t i;
		    for (i = 0; i < gene.size(); i++) {
		    
		     // affichage d'un fluent
		    fprintf(out," %s ", cptGetFluentName(cptGetFluentFromIndex(gene[i]))); 
			}
			fprintf(out," BKs=%d\n",bk);
		}
		void readFrom(istream& is)
	      {
		
			unsigned s;
    		is >> s;
    		
    		//cout<< "\n taille gene lu " << s ; 
    		gene.resize(s);
			 for (unsigned i=0; i<s; i++)
		      {
				unsigned uTmp;
				is >> uTmp;
				gene[i] = uTmp;
		      }				  
		      cout << this;
	      }
			
		bool notExist(size_t idfluent)
		{

			for(size_t i=0;i<gene.size();i++)
			{
				if(gene[i]==idfluent)return false;
			}
			return true;
		}	
		
		void sort(void)
		{
			std::sort(gene.begin(),gene.end());
		}
		
		/**
	 * @fn virtual ~eoCPTEO_V1()
	 * @brief destructeur de la classe
	 */
	  virtual ~Gene()
	  {
	    //gene.clear();
	  }	

};

/**
 * @class eoCPTEO_V1: public EO<FitT>
 * @file eoCPTEO_V1.h
 * @author Jacques Bibai
 * @brief definition d'un individu ou liste de gares dans la metaphore du TGV
 * Un eoCPTEO_V1 represente un individu dans la metaphore du TGV. 
 * Ici il est defini comme un vecteur de Gene. 
 * Chaque Gene representant une gare intermediaire dans la metaphore du TGV 
 */

template< class FitT>
class eoCPTEO_V1: public EO<FitT> 
{
	/**
	 * @brief definition des attributs de la classe template< class FitT> class eoCPTEO_V1: public EO<FitT> 
	 */
	private:
		// un individu est un vecteur d'entier. chaque entier repr$-3õ½sente l'indice d'un Fluent dans la table des fluents	
		std::vector<Gene> genome;
		
		// attribut permettant de sauvegarder un solution en terme de plan de l'individu
		CPTSolutionPlan plan;
		// attribut permettant de sauvegarder les variables temps de la solution
		double time,stime;
		// a retirer pour le tranporter vers la classe variable globale
		Gene init;
		long backtrack;
		long long Dplan;	
		size_t backGoal;
		size_t compressBk;
		size_t breakStation;
	public:
		
	  /**
		 * Constructeur vide
		 */
	  eoCPTEO_V1() 
	  { 
	  	plan=NULL;
	  }
	
	void setBreakStation(size_t i)
	{
		 breakStation=i;
	}
	 
	 size_t getBreakStation(void)
	 {
	 	return breakStation;	
	 }
	
	void setbackGoal(size_t b)
	{
		backGoal=b;
	}
	
	void setcompressbk(size_t b)
	{
		compressBk=b;
	}
	size_t getbackGoal()
	{
		return backGoal;
	}
	/**
	 * @fn std::vector<Gene> & getGenome()
	 * @brief methode permettant de recupere le vecteur decrivant un individu
	 * @return vecteur decrivant l'individu
	 */
	 std::vector<Gene> & getGenome()
	 {
	 	return genome;
	 }
	  std::vector<Gene> getGenome()const
	 {
	 	std::vector<Gene> tmp;
	 	tmp.resize(genome.size());
	 	for(size_t i=0;i<tmp.size();i++){
	 		tmp.push_back(genome[i]);
	 	}
	 	return tmp;
	 }
	 /**
	  * @fn void set(size_t, Gene&)
	  * @brief Cette methode permet de mettre a jour une gare : genome[position]=valeur
	  * @param position : numero de la gare
	  * @param valeur : reference vers une description
	  * 
	  */ 
	 void set(size_t position, Gene& valeur)
	 {
	 	genome[position].getGene().resize(valeur.size());
	 	for(size_t i=0;i<valeur.size();i++)
	 	genome[position].set(i,valeur.get(i));
	 	
	 }
	 /**
	  * @fn Gene& getInit()
	  * @brief cette methode permet de recuperer le gene decrivant l'etat initial
	  * @return une instance de la classe Gene decrivant la gare initiale
	  */ 
	 Gene& getInit(void)
	 {
	 	return init;
	 }
	
	 /**
	  * @fn void setInit(size_t,size_t)
	  * @brief cette methode permet la mise a jour d'un fluent (numero) de la gare initiale
	  * @param pos : position a mettre a jour
	  * @param val : nouvelle valeure  
	  */ 
	 void setInit(size_t pos,size_t val)
	 {
	 	init.set(pos,val);
	 }
	 /**
	  * @fn Gene& get(size_t)
	  * @brief accesseur : permet de recuperer une reference vers une station
	  * @param position : position de la station dont on cherche a recuperer la reference
	  * @return une instance de la classe Gene
	  */ 
	 Gene& get(size_t position)
	 {
	 	return genome[position];
	 } 
	 /**
	  * @fn void setPlan(CPTSolutionPlan)
	  * @brief mise a jour du plan decrivant les actions effectuees par l'individu de l'etat initial a l'etat final
	  * @param p : structure CPTSolutionPlan calculer par CPT
	  */ 
	 void setPlan(CPTSolutionPlan p)
	 {
	 	plan=p;
	 }
	 /**
	  * @fn void setPtime(double)
	  * @brief mise a jour de ptime
	  * @param t : double
	  */ 
	 void setDplan(long t)
	 {
	 	Dplan=t;
	 }
	 void setBacktrack(long b)
	 {
	 	backtrack=b;
	 }
	 /**
	  * @fn void setStime(double)
	  * @brief mise a jour de stime
	  * @param t : double
	  */ 
	 void setStime(double t)
	 {
	 	stime=t;
	 }
	 /**
	  * @fn void setTime(double)
	  * @brief mise a jour de Time
	  * @param t : double
	  */ 
	 void setTime(double t)
	 {
	 	time=t;
	 }
	 /**
	  * @fn CPTSolutionPlan getPlan()
	  * @brief accesseur : permet de recuperer le plan 
	  * @return une structure de type CPTSolutionPlan
	  */ 
	 CPTSolutionPlan getPlan()
	 {
	 	return plan;
	 }
	 /**
	  * @fn permute(eoCPTEO_V1&, size_t, size_t)
	  * @brief cette methode permet de permuter deux stations appartenant a deux individus.
	  * @param p : second individu
	  * @param pos1 : position de la station du second individu
	  * @param pos2 : position de la station du second individu
	  *  
	  */ 
	void permute(eoCPTEO_V1& p, size_t pos1,size_t pos2)
	{
			Gene tmp;
			tmp.getGene().resize(p.get(pos1).size());
			for(size_t i =0;i<tmp.size();i++){
				tmp.set(i,p.get(pos1).get(i));
			}
			
			p.set(pos1,genome[pos2]);
			set(pos2,tmp);
	}
	
	 
	 double mutexDistance(eoCPTEO_V1& p){
	 	double d=0;
	 	for(size_t i=0;i<getGenome().size();i++){
	 		int res=0;
     	if(i<p.getGenome().size())
     	for(size_t j=0;j<get(i).getGene().size();j++){
     		size_t k=0;
     		for(k=0;k<p.get(i).getGene().size();k++){
     			if(get(i).getGene().get(j)==p.get(i).getGene().get(k))break;
     		}
     		if(k==p.get(i).getGene().size())res++;
     				
     		}
     		d=d+(double)((int)get(i).getGene().size()-res)*(double)((int)get(i).getGene().size()-res);
     	}
    	 
	 	return d;
	 }
	/**
	 * @fn virtual ~eoCPTEO_V1()
	 * @brief destructeur de la classe
	 */
	  virtual ~eoCPTEO_V1()
	  {
	    //for(size_t i=0;i<genome.size();i++) genome[i].getGene().clear();
	    //genome.clear();
	  }
		
	/**
	 * @fn virtual string className() const 
	 */ 
	  virtual string className() const { return "eoCPTEO_V1"; }
	
	 /**
	  * @fn void printOn(ostream&)
	  * @brief methode permettant afficher un individu
	  */ 
	    void printOn(ostream& os) const
	    {
	      
	      EO<FitT>::printOn(os);
	      os << ' ';
	      os <<"\n" << genome.size();
	      os << "\nliste des stations\n";
	      size_t i;
	      
	      for(i=0;i<genome.size();i++){
	      	os  << " Station "<< i << genome[i] << "\n";
	      }
	      if(plan) 
	      {
	      	os << "BK " << backGoal <<" Solution trouvee : PLAN \n ";	 
	     	//if(plan) cptPrintSolutionPlan(stdout, plan); 
	     	os << "CompressionBK "<< compressBk << "\n Total Backtracks " << backtrack << "\n SearchTime " << stime << "\n TotalTime " << time << " \n PLAN " << Dplan << "\n";
	      }
	      else {
	      		os << "BK " << 0 <<" Pas de solution trouvee \n ";	 
	     	//if(plan) cptPrintSolutionPlan(stdout, plan); 
	     	os << "CompressionBK "<< 0 << "\n Total Backtracks " << 0 << "\n SearchTime " << 0 << "\n TotalTime " << 0 << " \n PLAN " << 0 << "\n";
	      }
	    }
	  
	 void printCPT(FILE* out)const
		{	fprintf(out,";liste des stations\n");
			size_t i;
		    for(i=0;i<genome.size();i++){
		    	fprintf(out,";Station%d ",i);
	      		genome[i].printCPT(out);
	     	}
	     	if(plan) 
	      {
	      	fprintf(out,";Goal BKs=%d\n",backGoal);
	      	fprintf(out,";Compression BKs=%d \n",compressBk);
	      }
		}   
	void printPlan()const
	    {
	    	if(plan) cptPrintSolutionPlan(stdout, plan); 
	    }
	    
	 	void printPlan(FILE* out)const
	    {	
	    	 fprintf(out,"Makespan : %lld\n", Dplan);
  			fprintf(out,"Backtracks : %ld\n",backtrack);
  			fprintf(out, "Total search time : %.2f\n",stime );
  			fprintf(out,"Preprocessing time : %.2f\n", time);
   	    	if(plan) cptPrintSolutionPlan(out, plan); 
	    }
	    
	    void printPlanVal(FILE* out)const
	    {	
	    	fprintf(out,";seed=%s \n",__seed.c_str());
	    	fprintf(out,";Makespan : %lld\n", Dplan);
  			fprintf(out,";Backtracks : %ld\n",backtrack);
  			fprintf(out, ";Total search time : %.2f\n",stime );
  			fprintf(out,";Preprocessing time : %.2f\n", time);
  			printCPT(out);
   	    	if(plan) cptPrintSolutionPlan2(out,plan,global.getPrecision()); 
  	    } 
	    void readFrom(istream& is)
	      {
			//cout << "****************************************************\n";
			EO<FitT>::readFrom(is);
			
			unsigned taille;
			is >> taille;
			//cout << " taille lu "<< taille << " \n";
			genome.resize(taille);
			std::string tmp;
			is >> tmp;
		//	cout << tmp.c_str();
			is >> tmp;
			//cout << tmp.c_str();
			is >> tmp;
		//	cout << tmp.c_str();
			//fin lecture entete
			for(size_t i=0;i<genome.size();i++)
			{
				is >> tmp;
				//cout << tmp.c_str();
				is >> taille;
				//cout <<" station " << taille ;
				Gene geneTmp;
				is >> taille;
				geneTmp.getGene().resize(taille);
				
				//cout << " taille du genome " << geneTmp.size();
				for(size_t j=0;j<geneTmp.size();j++)
				{
					is >> taille;
					geneTmp.set(j,taille);
					is >> tmp;
					cout << tmp.c_str();
					is >> tmp;
					cout << tmp.c_str();
					is >> tmp;
					
					cout << tmp.c_str();
				}
				//cout << geneTmp;
				
				set(i,geneTmp);
				is >> tmp;
				is >> taille;
				//cout << genome[i];
			}
			//lecture du nombre de BK solution finale
			double toto;
			is >> tmp;
			is >> toto;
			//solution trouvee : plan		
			is >> tmp;
			//cout << tmp.c_str();
			is >> tmp;
			//cout << tmp.c_str();
			is >> tmp;
			//cout << tmp.c_str();
			is >> tmp;
			//cout << tmp.c_str();
			is >> tmp;
			//cout << tmp.c_str();
			is >> taille;
			
			//total bk 
			is >> tmp;
			is >> tmp;
			is >> taille;
			//double toto;
	  	   size_t i;
	      for(i=0;i<3;i++){
	      	is >> tmp;
	      	is >> toto;
	      }
	   }

};

#endif /*GENOTYPE_H_*/
